from __future__ import annotations

from numpy import array, ndarray


class UGridNetwork1D:
    """This class is used for define/put/inquire/get Network1D data

    Attributes:
        name (str): The network name.
        node_x (ndarray): The x-coordinates of the network node.
        node_y (ndarray): The y-coordinates of the network node.
        edge_node (ndarray): The nodes defining each branch.
        edge_length (ndarray): The edge lengths.
        geometry_nodes_x (ndarray): The geometry nodes x coordinates.
        geometry_nodes_y (ndarray): The geometry nodes y coordinates.
        num_edge_geometry_nodes (ndarray): The number of geometry node on each branch.
        edge_order (ndarray): The order of the branches.
        node_id (str): The node names ids.
        node_name_long (str): The node long names.
        edge_id (list): The name of the branches.
        edge_long_name (list): The long name of the branches.
    """

    def __init__(
        self,
        name,
        node_x,
        node_y,
        edge_node,
        edge_length,
        geometry_nodes_x,
        geometry_nodes_y,
        num_edge_geometry_nodes,
        edge_order=array([]),
        node_id=[],
        node_name_long=[],
        edge_id=[],
        edge_long_name=[],
    ):
        self.name: str = name
        self.node_x: ndarray = node_x
        self.node_y: ndarray = node_y
        self.node_id: list = node_id
        self.node_long_name: list = node_name_long
        self.edge_node: ndarray = edge_node
        self.edge_length: ndarray = edge_length
        self.edge_order: ndarray = edge_order
        self.edge_id: list = edge_id
        self.edge_long_name: list = edge_long_name
        self.geometry_nodes_x: ndarray = geometry_nodes_x
        self.geometry_nodes_y: ndarray = geometry_nodes_y
        self.num_edge_geometry_nodes = num_edge_geometry_nodes
        self.is_spherical: bool = False
        self.start_index: int = 0


class UGridMesh1D:
    """This class is used for define/put/inquire/get Mesh1D data

    Attributes:
        name (c_char_p): The mesh name.
        network_name (c_char_p): The x-coordinates of the network node.
        node_x (ndarray):  The node x coordinate.
        node_y (ndarray):  The node y coordinate.
        edge_node (ndarray): The edge node connectivity.
        node_edge_id (ndarray):  The network edge id where every node lies.
        node_edge_offset (ndarray): The offset of each node on the network edge.
        node_name_id (list): A list of node names ids.
        node_name_long (c_char_p): A list of node long names.
        edge_edge_id (ndarray): The network edge id where every edge lies.
        edge_edge_offset (ndarray): The offset of each edge on the network edge.
        edge_x (ndarray): The edge x coordinate.
        edge_y (ndarray): The edge y coordinate.
        double_fill_value (c_double): The fill value for array of doubles.
        int_fill_value (c_int): The fill value for array of integers.
    """

    def __init__(
        self,
        name,
        network_name,
        node_edge_id,
        node_edge_offset,
        node_x=array([]),
        node_y=array([]),
        edge_node=array([]),
        edge_edge_id=array([]),
        edge_edge_offset=array([]),
        edge_x=array([]),
        edge_y=array([]),
        node_name_id=[],
        node_name_long=[],
        double_fill_value=-999.0,
        int_fill_value=-999,
    ):
        self.name: str = name
        self.network_name: str = network_name
        self.node_x: ndarray = node_x
        self.node_y: ndarray = node_y
        self.edge_node: ndarray = edge_node
        self.node_edge_id: ndarray = node_edge_id
        self.node_edge_offset: ndarray = node_edge_offset
        self.node_name_id: list = node_name_id
        self.node_name_long: list = node_name_long
        self.edge_edge_id: ndarray = edge_edge_id
        self.edge_edge_offset: ndarray = edge_edge_offset
        self.edge_x: ndarray = edge_x
        self.edge_y: ndarray = edge_y
        self.is_spherical: bool = False
        self.start_index: int = 0
        self.double_fill_value: float = double_fill_value
        self.int_fill_value: int = int_fill_value


class UGridMesh2D:
    """This class is used for define/put/inquire/get Mesh2D data

    Attributes:
        name (str): The mesh name.
        edge_nodes (ndarray): The nodes composing each mesh 2d edge.
        face_nodes (ndarray): The nodes composing each mesh 2d face.
        nodes_per_face (ndarray): The nodes composing each mesh 2d face.
        node_x (ndarray): The x-coordinates of the nodes.
        node_y (ndarray): The y-coordinates of the nodes.
        edge_x (ndarray): The x-coordinates of the mesh edges' middle points.
        edge_y (ndarray): The x-coordinates of the mesh edges' middle points.
        face_x (ndarray): The x-coordinates of the mesh faces' mass centers.
        face_y (ndarray): The y-coordinates of the mesh faces' mass centers.
        edge_face (ndarray): The edges composing each face.
        face_edge (ndarray): For each face, the edges composing it.
        face_face (ndarray): For each face, the neighboring faces.
        node_z (ndarray): The node z coordinates.
        edge_z (ndarray): The edge z coordinates.
        face_z (ndarray): The face z coordinates.
        layer_zs (ndarray): The z coordinates of a layer.
        interface_zs (ndarray): The z coordinates of a layer interface.
        boundary_node_connectivity (ndarray): To be detailed.
        volume_coordinates (ndarray): To be detailed.
        start_index (int): The start index used in arrays using indices, such as in the branch_node array.
        num_face_nodes_max (int): The maximum number of face nodes.
        is_spherical (c_int): 1 if coordinates are in a spherical system, 0 otherwise.
        double_fill_value (c_double): The fill value for array of doubles.
        int_fill_value (c_int): The fill value for array of integers.
    """

    def __init__(
        self,
        name,
        node_x,
        node_y,
        edge_node,
        face_node=array([]),
        edge_x=array([]),
        edge_y=array([]),
        face_x=array([]),
        face_y=array([]),
        edge_face=array([]),
        face_edge=array([]),
        face_face=array([]),
        node_z=array([]),
        edge_z=array([]),
        face_z=array([]),
        layer_zs=array([]),
        interface_zs=array([]),
        boundary_node_connectivity=array([]),
        volume_coordinates=array([]),
        start_index=0,
        num_face_nodes_max=4,
        is_spherical=False,
        double_fill_value=-999.0,
        int_fill_value=-999,
    ):
        self.name: str = name
        self.edge_node: ndarray = edge_node
        self.node_x: ndarray = node_x
        self.node_y: ndarray = node_y

        self.face_node: ndarray = face_node
        self.edge_x: ndarray = edge_x
        self.edge_y: ndarray = edge_y
        self.face_x: ndarray = face_x
        self.face_y: ndarray = face_y
        self.edge_face: ndarray = edge_face
        self.face_edge: ndarray = face_edge
        self.face_face: ndarray = face_face
        self.node_z: ndarray = node_z
        self.edge_z: ndarray = edge_z
        self.face_z: ndarray = face_z
        self.layer_zs: ndarray = layer_zs
        self.interface_zs: ndarray = interface_zs
        self.boundary_node_connectivity: ndarray = boundary_node_connectivity
        self.volume_coordinates: ndarray = volume_coordinates
        self.start_index: int = start_index
        self.num_face_nodes_max: int = num_face_nodes_max
        self.is_spherical: int = is_spherical
        self.double_fill_value: float = double_fill_value
        self.int_fill_value: int = int_fill_value


class UGridContacts:
    """This class is used for define/put/inquire/get Contacts data

    Attributes:
        name (str): The name of the contact entity.
        edges (ndarray): The actual contacts, expressed as pair of indices from a mesh index to another mesh index.
        contact_type (ndarray): For each contact its type.
        contact_name_id (list): The name of each contact.
        contact_name_long (list): The long name of each contact.
        mesh_from_name (str): The name of the mesh where the contacts start.
        mesh_to_name (str): The name of the mesh where the contacts ends.
        mesh_from_location (c_int): The location type (node, edge or face) at the contact start.
        mesh_to_location (c_int): The location type (node, edge or face) at the contact end.
        num_contacts (c_int): The number of contacts.
    """

    def __init__(
        self,
        name,
        edges,
        mesh_from_name,
        mesh_to_name,
        contact_type=array([]),
        contact_name_id=list,
        contact_name_long=list,
        mesh_from_location=0,
        mesh_to_location=0,
    ):
        self.name: str = name
        self.edges: ndarray = edges
        self.mesh_from_name: str = mesh_from_name
        self.mesh_to_name: str = mesh_to_name

        self.contact_type: ndarray = contact_type
        self.contact_name_id: list = contact_name_id
        self.contact_name_long: list = contact_name_long
        self.mesh_from_location: int = mesh_from_location
        self.mesh_to_location: int = mesh_to_location



class Error(Exception):
    """Base class for exceptions in this module."""


class InputError(Error):
    """Exception raised for errors in the input."""


class UGridError(Error):
    """Exception raised for errors coming from the MeshKernel library."""



import functools
import logging
import operator
import platform
from ctypes import CDLL, byref, c_char_p, c_int
from enum import IntEnum, unique
from pathlib import Path
from typing import Callable

import numpy as np
from meshkernel import Contacts, Mesh1d, Mesh2d
from numpy.ctypeslib import as_ctypes

from ugrid.c_structures import (
    CUGridContacts,
    CUGridMesh1D,
    CUGridMesh2D,
    CUGridNetwork1D,
    decode_byte_vector_to_list_of_strings,
    decode_byte_vector_to_string,
    numpy_array_to_ctypes,
)
from ugrid.errors import UGridError
from ugrid.py_structures import UGridContacts, UGridMesh1D, UGridMesh2D, UGridNetwork1D
from ugrid.version import __version__

logger = logging.getLogger(__name__)


@unique
class Status(IntEnum):
    SUCCESS = 0
    EXCEPTION = 1


class UGrid:
    """This class is the entry point for interacting with the UGridPy library"""

    def __init__(self, file_path, method):
        """Constructor of UGrid

        Raises:
            OSError: This gets raised in case UGrid is used within an unsupported OS.
        """

        # Determine OS
        system = platform.system()
        if system == "Windows":
            lib_path = Path(__file__).parent / "UGridApi.dll"
        elif system == "Linux":
            lib_path = Path(__file__).parent / "UGridApi.so"
        elif system == "Darwin":
            lib_path = Path(__file__).parent / "UGridApi.dylib"
        else:
            raise OSError(f"Unsupported operating system: {system}")

        self.lib = CDLL(str(lib_path))
        self.__open(file_path, method)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.__execute_function(self.lib.ug_file_close, self._file_id)
        error_message = self.__get_error()

        # Raise an exception if an error is present
        if error_message:
            raise UGridError(error_message)

    def __open(self, file_path: str, method: str) -> None:
        """Opens a NetCDF file containing a UGrid entities

        Comment

        Args:
            file_path (bool): The path of the file to open
            method (bool): The opening method ("r" for read, "w" for write, and "w+" for replace)
        """

        file_mode = c_int(0)
        if method == "r":
            self.__execute_function(self.lib.ug_file_read_mode, byref(file_mode))
        elif method == "w":
            self.__execute_function(self.lib.ug_file_write_mode, byref(file_mode))
        elif method == "w+":
            self.__execute_function(self.lib.ug_file_replace_mode, byref(file_mode))
        else:
            raise ValueError("Unsupported file mode")

        self._file_id = c_int(-1)

        file_path_bytes = bytes(file_path, encoding="utf8")
        self.__execute_function(
            self.lib.ug_file_open,
            file_path_bytes,
            file_mode,
            byref(self._file_id),
        )

    def __get_name_size(self):
        """Get the size of name strings"""
        name_size = c_int(0)
        self.__execute_function(self.lib.ug_name_get_length, byref(name_size))
        return name_size.value

    def __get_name_long_size(self):
        """Get the size of long name strings"""
        name_long_size = c_int(0)
        self.__execute_function(self.lib.ug_name_get_long_length, byref(name_long_size))
        return name_long_size.value

    def network1d_get_num_topologies(self) -> int:
        """Gets the number of network topologies contained in the file.

        Returns:
            int: The number of network topologies contained in the file.
        """

        topology_enum = self.topology_get_network1d_enum()

        topology_count = c_int(0)
        self.__execute_function(
            self.lib.ug_topology_get_count,
            self._file_id,
            c_int(topology_enum),
            byref(topology_count),
        )
        return topology_count.value

    def __network1d_inquire(self, topology_id) -> CUGridNetwork1D:
        """For internal use only.

        Inquires the network1d dimensions and names.

        Args:
            topology_id (int): The index of the network topology to inquire.

        Returns:
            CUGridNetwork1D: The network1d dimensions.
        """

        c_ugrid_network1d = CUGridNetwork1D()
        self.__execute_function(
            self.lib.ug_network1d_inq,
            self._file_id,
            c_int(topology_id),
            byref(c_ugrid_network1d),
        )
        return c_ugrid_network1d

    def network1d_get(self, topology_id) -> UGridNetwork1D:
        """Gets the network1d data.

        Args:
            topology_id (int): The index of the network1d topology to retrieve.

        Returns:
            UGridNetwork1D: The network1d (dimensions and data)
        """

        c_ugrid_network1d = self.__network1d_inquire(topology_id)
        name_size = self.__get_name_size()
        name_long_size = self.__get_name_long_size()

        ugrid_network1d = c_ugrid_network1d.allocate_memory(name_size, name_long_size)
        self.__execute_function(
            self.lib.ug_network1d_get,
            self._file_id,
            c_int(topology_id),
            byref(c_ugrid_network1d),
        )

        ugrid_network1d.is_spherical = bool(c_ugrid_network1d.is_spherical)
        ugrid_network1d.start_index = c_ugrid_network1d.start_index
        ugrid_network1d.name = decode_byte_vector_to_string(
            c_ugrid_network1d.name, name_size
        )
        ugrid_network1d.node_id = decode_byte_vector_to_list_of_strings(
            c_ugrid_network1d.node_id, c_ugrid_network1d.num_nodes, name_size
        )
        ugrid_network1d.node_long_name = decode_byte_vector_to_list_of_strings(
            c_ugrid_network1d.node_long_name,
            c_ugrid_network1d.num_nodes,
            name_long_size,
        )

        ugrid_network1d.edge_id = decode_byte_vector_to_list_of_strings(
            c_ugrid_network1d.edge_id,
            c_ugrid_network1d.num_edges,
            name_size,
        )
        ugrid_network1d.edge_long_name = decode_byte_vector_to_list_of_strings(
            c_ugrid_network1d.edge_long_name,
            c_ugrid_network1d.num_edges,
            name_long_size,
        )

        return ugrid_network1d

    def network1d_define(self, network1d: UGridNetwork1D) -> int:
        """Defines a new network1d in a UGrid file.

        Args:
            network1d (UGridNetwork1D): A network1d (dimensions and data)

        Returns:
            int: The index of the defined network topology.
        """

        name_size = self.__get_name_size()
        name_long_size = self.__get_name_long_size()

        c_ugrid_network = CUGridNetwork1D.from_py_structure(
            network1d, name_size, name_long_size
        )

        c_topology_id = c_int(-1)

        self.__execute_function(
            self.lib.ug_network1d_def,
            self._file_id,
            byref(c_ugrid_network),
            byref(c_topology_id),
        )

        return c_topology_id.value

    def network1d_put(self, topology_id: int, network1d: UGridNetwork1D) -> None:
        """Writes a new network1d to UGrid file.

        Args:
            topology_id (int): The index of the network1d topology to write.
            network1d (UGridNetwork1D): An instance of Network1D (with dimensions and data)
        """

        name_size = self.__get_name_size()
        name_long_size = self.__get_name_long_size()

        c_ugrid_network = CUGridNetwork1D.from_py_structure(
            network1d, name_size, name_long_size
        )

        self.__execute_function(
            self.lib.ug_network1d_put,
            self._file_id,
            c_int(topology_id),
            byref(c_ugrid_network),
        )

    def mesh1d_get_num_topologies(self) -> int:
        """Gets the number of mesh1d topologies contained in the file.

        Returns:
            int: The number of mesh1d topologies contained in the file.
        """

        topology_enum = self.topology_get_mesh1d_enum()

        topology_count = c_int(0)
        self.__execute_function(
            self.lib.ug_topology_get_count,
            self._file_id,
            c_int(topology_enum),
            byref(topology_count),
        )
        return topology_count.value

    def __mesh1d_inquire(self, topology_id) -> CUGridMesh1D:
        """For internal use only.

        Inquires the mesh1d dimensions and names.

        Args:
            topology_id (int): The index of the mesh1d topology to inquire.

        Returns:
            CUGridMesh1D: The mesh1d dimensions.

        """

        c_ugrid_mesh1d = CUGridMesh1D()
        self.__execute_function(
            self.lib.ug_mesh1d_inq,
            self._file_id,
            c_int(topology_id),
            byref(c_ugrid_mesh1d),
        )
        return c_ugrid_mesh1d

    def mesh1d_get(self, topology_id) -> UGridMesh1D:
        """Gets the mesh1d data.

        Args:
            topology_id (int): The index of the mesh1d topology to retrieve.

        Returns:
            UGridMesh1D: The mesh1d (dimensions and data)
        """

        c_mesh1d = self.__mesh1d_inquire(topology_id)
        name_size = self.__get_name_size()
        name_long_size = self.__get_name_long_size()

        ugrid_mesh1d = c_mesh1d.allocate_memory(name_size, name_long_size)

        self.__execute_function(
            self.lib.ug_mesh1d_get,
            self._file_id,
            c_int(topology_id),
            byref(c_mesh1d),
        )

        ugrid_mesh1d.is_spherical = bool(c_mesh1d.is_spherical)
        ugrid_mesh1d.start_index = c_mesh1d.start_index
        ugrid_mesh1d.double_fill_value = c_mesh1d.double_fill_value
        ugrid_mesh1d.int_fill_value = c_mesh1d.int_fill_value

        ugrid_mesh1d.name = decode_byte_vector_to_string(c_mesh1d.name, name_size)
        ugrid_mesh1d.network_name = decode_byte_vector_to_string(
            c_mesh1d.network_name, name_size
        )

        ugrid_mesh1d.node_name_id = decode_byte_vector_to_list_of_strings(
            c_mesh1d.node_name_id, c_mesh1d.num_nodes, name_size
        )
        ugrid_mesh1d.node_name_long = decode_byte_vector_to_list_of_strings(
            c_mesh1d.node_name_long, c_mesh1d.num_nodes, name_long_size
        )

        return ugrid_mesh1d

    def mesh1d_define(self, mesh1d: UGridMesh1D) -> int:
        """Defines a new mesh1d in a UGrid file.

        Args:
            mesh1d (UGridMesh1D): A mesh1d (dimensions and data)

        Returns:
            int: The index of the defined mesh1d topology.
        """

        name_size = self.__get_name_size()
        name_long_size = self.__get_name_long_size()

        c_ugrid_mesh1d = CUGridMesh1D.from_py_structure(
            mesh1d, name_size, name_long_size
        )

        c_topology_id = c_int(-1)

        self.__execute_function(
            self.lib.ug_mesh1d_def,
            self._file_id,
            byref(c_ugrid_mesh1d),
            byref(c_topology_id),
        )

        return c_topology_id.value

    def mesh1d_put(self, topology_id: int, mesh1d: UGridMesh1D) -> None:
        """Writes a new mesh1d to UGrid file.

        Args:
            topology_id (int): The index of the mesh1d topology to write.
            mesh1d (UGridMesh1D): An instance of mesh1d (with dimensions and data)
        """

        name_size = self.__get_name_size()
        name_long_size = self.__get_name_long_size()

        c_ugrid_mesh1d = CUGridMesh1D.from_py_structure(
            mesh1d, name_size, name_long_size
        )

        self.__execute_function(
            self.lib.ug_mesh1d_put,
            self._file_id,
            c_int(topology_id),
            byref(c_ugrid_mesh1d),
        )

    def mesh2d_get_num_topologies(self) -> int:
        """Description

        Gets the number of mesh2d topologies contained in the file.

        Returns:
            int: The number of mesh2d topologies contained in the file.

        """

        topology_enum = self.topology_get_mesh2d_enum()

        topology_count = c_int(0)
        self.__execute_function(
            self.lib.ug_topology_get_count,
            self._file_id,
            c_int(topology_enum),
            byref(topology_count),
        )
        return topology_count.value

    def __mesh2d_inquire(self, topology_id) -> CUGridMesh2D:
        """For internal use only.

        Inquires the mesh2d dimensions and names.

        Args:
            topology_id (int): The index of the mesh2d topology to inquire.

        Returns:
            CUGridMesh2D: The mesh2d dimensions.

        """

        c_ugrid_mesh2d = CUGridMesh2D()
        self.__execute_function(
            self.lib.ug_mesh2d_inq,
            self._file_id,
            c_int(topology_id),
            byref(c_ugrid_mesh2d),
        )
        return c_ugrid_mesh2d

    def mesh2d_get(self, topology_id) -> UGridMesh2D:
        """Gets the mesh2d data.

        Args:
            topology_id (int): The index of the mesh2d topology to retrieve.

        Returns:
            UGridMesh2D: The mesh2d (dimensions and data)
        """

        c_ugrid_mesh2d = self.__mesh2d_inquire(topology_id)
        long_name_size = self.__get_name_long_size()

        ugrid_mesh2d = c_ugrid_mesh2d.allocate_memory(long_name_size)
        self.__execute_function(
            self.lib.ug_mesh2d_get,
            self._file_id,
            c_int(topology_id),
            byref(c_ugrid_mesh2d),
        )

        ugrid_mesh2d.is_spherical = bool(c_ugrid_mesh2d.is_spherical)
        ugrid_mesh2d.start_index = c_ugrid_mesh2d.start_index
        ugrid_mesh2d.double_fill_value = c_ugrid_mesh2d.double_fill_value
        ugrid_mesh2d.int_fill_value = c_ugrid_mesh2d.int_fill_value

        ugrid_mesh2d.name = decode_byte_vector_to_string(
            c_ugrid_mesh2d.name, long_name_size
        )

        return ugrid_mesh2d

    def mesh2d_define(self, ugrid_mesh2d: UGridMesh2D) -> int:
        """Defines a new mesh2d in a UGrid file.

        Args:
            ugrid_mesh2d (UGridMesh2D): A mesh2d (dimensions and data)

        Returns:
            int: The index of the defined mesh2d topology.
        """

        long_name_size = self.__get_name_long_size()

        c_ugrid_mesh2d = CUGridMesh2D.from_py_structure(ugrid_mesh2d, long_name_size)

        c_topology_id = c_int(-1)

        self.__execute_function(
            self.lib.ug_mesh2d_def,
            self._file_id,
            byref(c_ugrid_mesh2d),
            byref(c_topology_id),
        )

        return c_topology_id.value

    def mesh2d_put(self, topology_id: int, ugrid_mesh2d: UGridMesh2D) -> None:
        """Writes a new mesh2d in a UGrid file.

        Args:
            topology_id (int): The index of the mesh2d topology to write.
            ugrid_mesh2d (UGridMesh2D): A mesh2d (dimensions and data)
        """

        name_size = self.__get_name_size()
        c_ugrid_mesh2d = CUGridMesh2D.from_py_structure(ugrid_mesh2d, name_size)

        self.__execute_function(
            self.lib.ug_mesh2d_put,
            self._file_id,
            c_int(topology_id),
            byref(c_ugrid_mesh2d),
        )

    @staticmethod
    def from_meshkernel_mesh2d_to_ugrid_mesh2d(
        mesh2d: Mesh2d, name: str, is_spherical: bool
    ) -> UGridMesh2D:
        """Converts a meshkernel mesh2d to ugrid mesh2d

        Args:
            mesh2d (Mesh2d): An instance of a meshkernel mesh2d
            name (str): The name of mesh2d
            is_spherical (bool): Spherical or cartesian coordinate system
        """

        num_faces = len(mesh2d.nodes_per_face)
        if num_faces > 0:
            num_face_nodes_max = np.max(mesh2d.nodes_per_face)
            face_nodes_array = np.full(
                num_faces * num_face_nodes_max, dtype=np.int, fill_value=-1
            )
            for index, (face_nodes, nodes_per_face) in enumerate(
                zip(mesh2d.face_nodes, mesh2d.nodes_per_face)
            ):
                face_nodes_array[
                    index * num_face_nodes_max : index * num_face_nodes_max
                    + nodes_per_face
                ] = face_nodes

            return UGridMesh2D(
                name=name,
                node_x=mesh2d.node_x,
                node_y=mesh2d.node_y,
                edge_node=mesh2d.edge_nodes,
                face_node=face_nodes_array,
                edge_x=mesh2d.edge_x,
                edge_y=mesh2d.edge_y,
                face_x=mesh2d.face_x,
                face_y=mesh2d.face_y,
                num_face_nodes_max=num_face_nodes_max,
                is_spherical=is_spherical,
                double_fill_value=-999.0,
                int_fill_value=-999,
            )

        return UGridMesh2D(
            name=name,
            node_x=mesh2d.node_x,
            node_y=mesh2d.node_y,
            edge_node=mesh2d.edge_nodes,
        )

    @staticmethod
    def from_meshkernel_mesh1d_to_ugrid_mesh1d(
        mesh1d: Mesh1d,
        name: str,
        network_name: str,
        node_edge_id: np.ndarray,
        node_edge_offset: np.ndarray,
        node_name_id: list,
        node_name_long: list,
        edge_edge_id: np.ndarray,
        edge_edge_offset: np.ndarray,
        edge_x: np.ndarray,
        edge_y: np.ndarray,
        double_fill_value: float,
        int_fill_value: int,
    ) -> UGridMesh1D:
        """Converts a meshkernel mesh1d to a ugrid mesh1d

        Args:
            mesh1d (Mesh1d): An instance of a meshkernel mesh1d
            network_name (c_char_p): The x-coordinates of the network node.
            node_x (ndarray):  The node x coordinate.
            node_y (ndarray):  The node y coordinate.
            node_edge_id (ndarray):  The network edge id where every node lies.
            node_edge_offset (ndarray): The offset of each node on the network edge
            node_name_id (list): A list of node names ids.
            node_name_long (c_char_p): A list of node long names.
            edge_edge_id (ndarray): The network edge id where every edge lies.
            edge_edge_offset (ndarray): The offset of each edge on the network edge.
            edge_x (ndarray): The edge x coordinate.
            edge_y (ndarray): The edge y coordinate.
            double_fill_value (c_double): The fill value for array of doubles.
            int_fill_value (c_int): The fill value for array of integers.
        """

        ugrid_mesh1d = UGridMesh1D(
            name=name,
            network_name=network_name,
            node_x=mesh1d.node_x,
            node_y=mesh1d.node_y,
            edge_node=mesh1d.edge_nodes,
            node_edge_id=node_edge_id,
            node_edge_offset=node_edge_offset,
            node_name_id=node_name_id,
            node_name_long=node_name_long,
            edge_edge_id=edge_edge_id,
            edge_edge_offset=edge_edge_offset,
            edge_x=edge_x,
            edge_y=edge_y,
            double_fill_value=double_fill_value,
            int_fill_value=int_fill_value,
        )

        return ugrid_mesh1d

    @staticmethod
    def from_meshkernel_contacts_to_ugrid_contacts(
        contacts: Contacts,
        name: str,
        contact_type: np.ndarray,
        contact_name_id: str,
        contact_name_long: str,
        mesh_from_name: str,
        mesh_to_name: str,
        mesh_from_location: int,
        mesh_to_location: int,
    ) -> UGridContacts:
        """Converts a meshkernel mesh1d to a ugrid mesh1d

        Args:
            contacts (Contacts): An instance of a meshkernel contacts entity
            name (str): The name of the contact entity.
            contact_type (ndarray): For each contact its type.
            contact_name_id (list): The name of each contact.
            contact_name_long (list): The long name of each contact.
            mesh_from_name (str): The name of the mesh where the contacts start.
            mesh_to_name (str): The name of the mesh where the contacts ends.
            mesh_from_location (c_int): The location type (node, edge or face) at the contact start.
            mesh_to_location (c_int): The location type (node, edge or face) at the contact end.
        """

        num_edges = len(contacts.mesh1d_indices)
        edges_array = np.full(num_edges * 2, dtype=np.int, fill_value=-1)
        for index, (mesh1d_indices, mesh2d_indices) in enumerate(
            zip(contacts.mesh1d_indices, contacts.mesh2d_indices)
        ):
            edges_array[index * 2] = mesh1d_indices
            edges_array[index * 2 + 1] = mesh2d_indices

        ugrid_contacts = UGridContacts(
            name=name,
            edges=edges_array,
            mesh_from_name=mesh_from_name,
            mesh_to_name=mesh_to_name,
            contact_type=contact_type,
            contact_name_id=contact_name_id,
            contact_name_long=contact_name_long,
            mesh_from_location=mesh_from_location,
            mesh_to_location=mesh_to_location,
        )

        return ugrid_contacts

    def contacts_get_num_topologies(self) -> int:
        """Description

        Gets the number of contacts topologies contained in the file.

        Returns:
            int: The number of contacts topologies contained in the file.

        """

        topology_enum = self.topology_get_contacts_enum()

        topology_count = c_int(0)
        self.__execute_function(
            self.lib.ug_topology_get_count,
            self._file_id,
            c_int(topology_enum),
            byref(topology_count),
        )
        return topology_count.value

    def __contacts_inquire(self, topology_id) -> CUGridContacts:
        """For internal use only.

        Inquires the contacts dimensions and names.

        Args:
            topology_id (int): The index of the contacts topology to inquire.

        Returns:
            CUGridContacts: The contacts dimensions.

        """

        c_ugrid_contacts = CUGridContacts()
        self.__execute_function(
            self.lib.ug_contacts_inq,
            self._file_id,
            c_int(topology_id),
            byref(c_ugrid_contacts),
        )
        return c_ugrid_contacts

    def contacts_get(self, topology_id) -> UGridContacts:
        """Gets the contacts data.

        Args:
            topology_id (int): The index of the contacts topology to retrieve.

        Returns:
            UGridContacts: The contacts (dimensions and data)
        """

        c_ugrid_contacts = self.__contacts_inquire(topology_id)
        name_size = self.__get_name_size()
        name_long_size = self.__get_name_long_size()

        ugrid_contacts = c_ugrid_contacts.allocate_memory(name_size, name_long_size)
        self.__execute_function(
            self.lib.ug_contacts_get,
            self._file_id,
            c_int(topology_id),
            byref(c_ugrid_contacts),
        )

        ugrid_contacts.name = decode_byte_vector_to_string(
            c_ugrid_contacts.name, name_size
        )
        ugrid_contacts.contact_name_id = decode_byte_vector_to_list_of_strings(
            c_ugrid_contacts.contact_name_id,
            c_ugrid_contacts.num_contacts,
            name_size,
        )
        ugrid_contacts.contact_name_long = decode_byte_vector_to_list_of_strings(
            c_ugrid_contacts.contact_name_long,
            c_ugrid_contacts.num_contacts,
            name_long_size,
        )

        ugrid_contacts.mesh_from_name = decode_byte_vector_to_string(
            c_ugrid_contacts.mesh_from_name, name_size
        )
        ugrid_contacts.mesh_to_name = decode_byte_vector_to_string(
            c_ugrid_contacts.mesh_to_name, name_size
        )

        return ugrid_contacts

    def contacts_define(self, contacts: UGridContacts) -> int:
        """Defines a new contacts in a UGrid file.

        Args:
            contacts (UGridContacts): An instance of Contacts class (dimensions and data)

        Returns:
            int: The index of the defined contacts topology.
        """

        name_size = self.__get_name_size()
        name_long_size = self.__get_name_long_size()

        c_ugrid_contacts = CUGridContacts.from_py_structure(
            contacts, name_size, name_long_size
        )

        c_topology_id = c_int(-1)

        self.__execute_function(
            self.lib.ug_contacts_def,
            self._file_id,
            byref(c_ugrid_contacts),
            byref(c_topology_id),
        )

        return c_topology_id.value

    def contacts_put(self, topology_id: int, contacts: UGridContacts) -> None:
        """Writes a new contacts in a UGrid file.

        Args:
            topology_id (int): The index of the contacts topology to write.
            contacts (UGridContacts): A contacts (dimensions and data)
        """
        name_size = self.__get_name_size()
        name_long_size = self.__get_name_long_size()

        c_ugrid_contacts = CUGridContacts.from_py_structure(
            contacts, name_size, name_long_size
        )

        self.__execute_function(
            self.lib.ug_contacts_put,
            self._file_id,
            c_int(topology_id),
            byref(c_ugrid_contacts),
        )

    def __get_error(self) -> str:
        c_error_message = c_char_p()
        self.__execute_function(self.lib.ug_error_get, byref(c_error_message))
        return c_error_message.value.decode("ASCII")

    def get_ugrid_version(self) -> str:
        """Get the version of the C++ UGrid library

        Returns:
            str: The version string
        """

        c_ugrid_version = c_char_p()
        self.__execute_function(self.lib.mkernel_get_version, byref(c_ugrid_version))
        return c_ugrid_version.value.decode("ASCII")

    def get_ugridpy_version(self) -> str:
        """Get the version of this Python wrapper

        Returns:
            str: The version string
        """

        return __version__

    def __execute_function(self, function: Callable, *args):
        """Utility function to execute a C function of UGrid and checks its status.

        Args:
            function (Callable): The function which we want to call.
            args: Arguments which will be passed to `function`.

        Raises:
            UGridError: This exception gets raised,
             if the UGrid library reports an error.
        """
        if function(*args) != Status.SUCCESS:
            error_message = self.__get_error()
            raise UGridError(error_message)

    def entity_get_node_location_enum(self) -> int:
        """Get the node location enum value

        Returns:
            int: The node location enum value
        """

        location = c_int(0)
        self.__execute_function(
            self.lib.ug_entity_get_node_location_enum, byref(location)
        )
        return location.value

    def entity_get_edge_location_enum(self) -> int:
        """Get the edge location enum value

        Returns:
            int: The edge location enum value
        """

        location = c_int(0)
        self.__execute_function(
            self.lib.ug_entity_get_edge_location_enum, byref(location)
        )
        return location.value

    def entity_get_face_location_enum(self) -> int:
        """Get the face location enum value

        Returns:
            int: The face location enum value
        """

        location = c_int(0)
        self.__execute_function(
            self.lib.ug_entity_get_face_location_enum, byref(location)
        )
        return location.value

    def topology_get_network1d_enum(self) -> int:

        """Gets the topology enum value associated with network1d.

        Returns:
            int: the topology enum value associated with network1d.
        """
        topology_enum = c_int(0)
        self.__execute_function(
            self.lib.ug_topology_get_network1d_enum, byref(topology_enum)
        )

        return topology_enum.value

    def topology_get_mesh1d_enum(self) -> int:

        """Gets the topology enum value associated with mesh1d.

        Returns:
            int: the topology enum value associated with mesh1d.
        """
        topology_enum = c_int(0)
        self.__execute_function(
            self.lib.ug_topology_get_mesh1d_enum, byref(topology_enum)
        )

        return topology_enum.value

    def topology_get_mesh2d_enum(self) -> int:

        """Gets the topology enum value associated with mesh2d.

        Returns:
            int: the topology enum value associated with mesh2d.
        """
        topology_enum = c_int(0)
        self.__execute_function(
            self.lib.ug_topology_get_mesh2d_enum, byref(topology_enum)
        )

        return topology_enum.value

    def topology_get_contacts_enum(self) -> int:

        """Gets the topology enum value associated with contacts.

        Returns:
            int: the topology enum value associated with contacts.
        """
        topology_enum = c_int(0)
        self.__execute_function(
            self.lib.ug_topology_get_contacts_enum, byref(topology_enum)
        )

        return topology_enum.value

    def __topology_count_data_variables(
        self, topology_id: int, topology_type: int, location: int
    ) -> int:
        """Gets the number of data variables for a specific topology at a specific location

        Args:
            topology_id (int): The index of the topology type.
            topology_type (int): The list with the attributes values.
            location (int): The location type

        Returns:
            int: The number of data
        """

        data_variable_count = c_int(0)
        self.__execute_function(
            self.lib.ug_topology_count_data_variables,
            self._file_id,
            c_int(topology_type),
            c_int(topology_id),
            c_int(location),
            byref(data_variable_count),
        )

        return data_variable_count.value

    def topology_get_data_variables(
        self, topology_id: int, topology_type: int, location: int
    ) -> list:
        """Gets the data variables for a specific topology at a specific location

        Args:
            topology_id (int): The index of the topology type.
            topology_type (int): The list with the attributes values.
            location (int): The location type

        Returns:
            list: The list of data variables
        """

        num_data_variables = self.__topology_count_data_variables(
            topology_id, topology_type, location
        )
        name_long_size = self.__get_name_long_size()

        buffer_size = name_long_size * num_data_variables
        string_buffer = " " * buffer_size

        string_buffer_encoded = c_char_p(string_buffer.encode("ASCII"))
        self.__execute_function(
            self.lib.ug_topology_count_data_variables,
            self._file_id,
            c_int(topology_type),
            c_int(topology_id),
            c_int(location),
            string_buffer_encoded,
        )

        attribute_list = decode_byte_vector_to_list_of_strings(
            string_buffer_encoded.value, num_data_variables, name_long_size
        )

        return attribute_list.value

    def __adjust_name(self, name: str) -> str:
        long_name = self.__get_name_long_size()
        name_long = name.ljust(long_name)
        return name_long

    def __variable_count_attributes(self, variable_name: str) -> int:
        """Counts the number of variable attributes.

        Args:
            variable_name (str): The variable name.

        Returns:
            int: The number of attributes
        """

        attributes_count = c_int(0)
        variable_name_long = self.__adjust_name(variable_name)
        c_variable_name_encoded = c_char_p(variable_name_long.encode("ASCII"))
        self.__execute_function(
            self.lib.ug_variable_count_attributes,
            self._file_id,
            c_variable_name_encoded,
            byref(attributes_count),
        )
        return attributes_count.value

    def variable_get_attributes_names(self, variable_name: str) -> list:
        """Gets the variable attribute names.

        Args:
            variable_name (str): The variable name.

        Returns:
            list: A list containing the variable attribute names.
        """

        num_attributes = self.__variable_count_attributes(variable_name)

        name_long_size = self.__get_name_long_size()
        buffer_size = name_long_size * num_attributes
        attribute_buffer = " " * buffer_size
        attribute_buffer_encoded = c_char_p(attribute_buffer.encode("ASCII"))

        variable_name_long = self.__adjust_name(variable_name)
        c_variable_name_encoded = c_char_p(variable_name_long.encode("ASCII"))

        self.__execute_function(
            self.lib.ug_variable_get_attributes_names,
            self._file_id,
            c_variable_name_encoded,
            attribute_buffer_encoded,
        )

        attribute_list = decode_byte_vector_to_list_of_strings(
            attribute_buffer_encoded.value, num_attributes, name_long_size
        )

        return attribute_list

    def variable_get_attributes_values(self, variable_name: str) -> list:
        """Gets the variable attribute values.

        Args:
            variable_name (str): The variable name.

        Returns:
            list: A list containing the variable attribute values.
        """

        num_attributes = self.__variable_count_attributes(variable_name)
        name_long_size = self.__get_name_long_size()

        buffer_size = name_long_size * num_attributes
        string_buffer = " " * buffer_size
        string_buffer_encoded = c_char_p(string_buffer.encode("ASCII"))

        variable_name_long = self.__adjust_name(variable_name)
        c_variable_name_encoded = c_char_p(variable_name_long.encode("ASCII"))

        self.__execute_function(
            self.lib.ug_variable_get_attributes_values,
            self._file_id,
            c_variable_name_encoded,
            string_buffer_encoded,
        )

        attribute_list = decode_byte_vector_to_list_of_strings(
            string_buffer_encoded.value, num_attributes, name_long_size
        )

        return attribute_list

    def __variable_get_dimensions(self, variable_name) -> np.ndarray:
        """Gets the variable dimensions

        Args:
            data_variable_name (str): The data variable name.

        Returns:
            np.ndarray: An array with the dimension values
        """

        variable_name_long = self.__adjust_name(variable_name)
        c_variable_name_encoded = c_char_p(variable_name_long.encode("ASCII"))
        c_num_dimensions = c_int(0)
        self.__execute_function(
            self.lib.ug_variable_count_dimensions,
            self._file_id,
            c_variable_name_encoded,
            byref(c_num_dimensions),
        )

        dimension_vec = np.empty(c_num_dimensions.value, dtype=np.int)
        dimension_vec_ptr = as_ctypes(dimension_vec)
        self.__execute_function(
            self.lib.ug_variable_get_data_dimensions,
            self._file_id,
            c_variable_name_encoded,
            dimension_vec_ptr,
        )
        return dimension_vec

    def variable_get_data_double(self, variable_name: str) -> np.ndarray:
        """Gets the variable data as a flat array of double

        Args:
            variable_name (str): The variable name.

        Returns:
            np.ndarray: A numpy array with the variable data
        """

        dimension_vec = self.__variable_get_dimensions(variable_name)
        data_vec_dimension = functools.reduce(operator.mul, dimension_vec)

        data_vec = np.empty(data_vec_dimension, dtype=np.double)
        data_vec_ptr = as_ctypes(data_vec)
        variable_name_long = self.__adjust_name(variable_name)
        c_variable_name_encoded = c_char_p(variable_name_long.encode("ASCII"))
        self.__execute_function(
            self.lib.ug_variable_get_data_double,
            self._file_id,
            c_variable_name_encoded,
            data_vec_ptr,
        )

        return data_vec

    def variable_get_data_int(self, variable_name: str) -> np.ndarray:
        """Gets the variable data as a flat array of strings

        Args:
            variable_name (str): The variable name.

        Returns:
            np.ndarray: A numpy array with the variable data
        """

        dimension_vec = self.__variable_get_dimensions(variable_name)

        data_vec_dimension = functools.reduce(operator.mul, dimension_vec)

        data_vec = np.empty(data_vec_dimension, dtype=np.int)
        data_vec_ptr = as_ctypes(data_vec)
        variable_name_long = self.__adjust_name(variable_name)
        c_variable_name_encoded = c_char_p(variable_name_long.encode("ASCII"))
        self.__execute_function(
            self.lib.ug_variable_get_data_int,
            self._file_id,
            c_variable_name_encoded,
            data_vec_ptr,
        )

        return data_vec

    def __variable_int_define(self, variable_name: str):
        """Defines a new integer variable

        Args:
            variable_name (str): The variable name.
        """

        variable_name_long = self.__adjust_name(variable_name)
        c_variable_name_encoded = c_char_p(variable_name_long.encode("ASCII"))

        self.__execute_function(
            self.lib.ug_variable_int_define, self._file_id, c_variable_name_encoded
        )

    def __attribute_define(
        self, variable_name: str, attribute_name: str, attribute_values
    ):
        """Defines a new variable attribute with a name and an array of values.

        Args:
            variable_name (str): The variable name.
            attribute_name (str): The attribute name.
            attribute_values: The attribute values. Can be a numpy array of int, floats or a string
        """

        variable_name_long = self.__adjust_name(variable_name)
        c_variable_name_encoded = c_char_p(variable_name_long.encode("ASCII"))

        attribute_name_long = self.__adjust_name(attribute_name)
        c_attribute_name_encoded = c_char_p(attribute_name_long.encode("ASCII"))

        attribute_values_len = c_int(len(attribute_values))

        if isinstance(
            attribute_values, np.ndarray
        ) and attribute_values.dtype is np.dtype("int32"):
            funct = self.lib.ug_attribute_int_define
            c_attribute_values = numpy_array_to_ctypes(attribute_values)
        elif isinstance(
            attribute_values, np.ndarray
        ) and attribute_values.dtype is np.dtype("float"):
            funct = self.lib.ug_attribute_double_define
            c_attribute_values = numpy_array_to_ctypes(attribute_values)
        elif isinstance(attribute_values, str):
            funct = self.lib.ug_attribute_char_define
            c_attribute_values = c_char_p(attribute_values.encode("ASCII"))
        else:
            raise UGridError(
                "__attribute_define: attribute_values types are not supported"
            )

        self.__execute_function(
            funct,
            self._file_id,
            c_variable_name_encoded,
            c_attribute_name_encoded,
            c_attribute_values,
            attribute_values_len,
        )

    def variable_int_with_attributes_define(
        self, variable_name: str, variable_dict: dict
    ):
        """Defines a new variable with attributes. It can be used to write the coordinate reference system to file.

        Args:
            variable_name (str): The new variable name.
            variable_dict (dict): A dictionary containing the variable values and names
        """

        self.__variable_int_define(variable_name)
        for names, values in variable_dict.items():
            self.__attribute_define(variable_name, names, values)

    def attribute_global_define(self, variable_dict: dict):
        """Defines new global attributes

        Args:
            variable_dict (dict): A dictionary containing the attribute names and values.
        """
        for attribute_name, attribute_value in variable_dict.items():

            attribute_name_long = self.__adjust_name(attribute_name)
            c_attribute_name_encoded = c_char_p(attribute_name_long.encode("ASCII"))

            c_attribute_value_encoded = c_char_p(attribute_value.encode("ASCII"))
            attribute_values_len = c_int(len(attribute_value))

            self.__execute_function(
                self.lib.ug_attribute_global_char_define,
                self._file_id,
                c_attribute_name_encoded,
                c_attribute_value_encoded,
                attribute_values_len,
            )



import codecs
import os.path
import platform

from setuptools import setup

author_dict = {
    "Luca Carniato": "luca.carniato@deltares.nl",
}
__author__ = ", ".join(author_dict.keys())
__author_email__ = ", ".join(s for _, s in author_dict.items())


def read(rel_path: str) -> str:
    """Used to read a text file

    Args:
        rel_path (str): Relative path to the file

    Returns:
        str: File content
    """
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path: str) -> str:
    """Get the version string

    Args:
        rel_path (str): Relative path to the file

    Raises:
        RuntimeError: Raised if the version string could not be found

    Returns:
        str: The version string
    """
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]

    raise RuntimeError("Unable to find version string.")


def get_runtime_libraries() -> str:
    """Get the UGrid runtime libraries

    Raises:
        OSError: If the operating system is not supported

    Returns:
        str: Filename of the MeshKernel library
    """
    system = platform.system()
    if system == "Windows":
        return "*.dll"
    elif system == "Linux":
        return "*.so"
    elif system == "Darwin":
        return "*.dylib"
    else:
        raise OSError(f"Unsupported operating system: {system}")


try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

    class bdist_wheel(_bdist_wheel):
        """Class describing our wheel.
        Basically it says that it is not a pure Python package,
        but it also does not contain any Python source and
        therefore works for all Python versions
        """

        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            # Mark us as not a pure python package
            self.root_is_pure = False

        def get_tag(self):
            python, abi, plat = _bdist_wheel.get_tag(self)
            # We don't contain any python source
            python, abi = "py3", "none"
            return python, abi, plat


except ImportError:
    bdist_wheel = None


long_description = read("README.md")

setup(
    name="UGrid",
    description="`UGrid` is a library to write and read UGrid files.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=__author__,
    author_email=__author_email__,
    url="https://github.com/Deltares/UGridPy",
    license="MIT",
    platforms="Windows, Linux",
    install_requires=["numpy", "meshkernel"],
    extras_require={
        "tests": ["pytest", "pytest-cov", "nbval"],
        "lint": [
            "flake8",
            "black==21.4b1",
            "isort",
        ],
        "docs": ["sphinx", "sphinx_book_theme", "myst_nb"],
    },
    python_requires=">=3.8",
    packages=["ugrid"],
    package_data={
        "ugrid": [get_runtime_libraries()],
    },
    cmdclass={"bdist_wheel": bdist_wheel},
    version=get_version("ugrid/version.py"),
    classifiers=["Topic :: Scientific/Engineering :: Mathematics"],
)



# If you change these imports,
# do not forget to sync the docs at "docs/api"
from ugrid.errors import InputError, UGridError
from ugrid.py_structures import UGridContacts, UGridMesh1D, UGridMesh2D, UGridNetwork1D
from ugrid.ugrid import UGrid
from ugrid.version import __version__



from __future__ import annotations

from ctypes import POINTER, Structure, c_char_p, c_double, c_int

import numpy as np
from numpy.ctypeslib import as_ctypes

from ugrid.py_structures import UGridContacts, UGridMesh1D, UGridMesh2D, UGridNetwork1D


def decode_byte_vector_to_string(byte_vector: bytes, ncolumns: int) -> str:
    """From byte vector to string"""
    return byte_vector[:ncolumns].decode("ASCII").strip()


def decode_byte_vector_to_list_of_strings(
    byte_vector: bytes, nrows: int, str_size: int
) -> list:
    """From byte vector to a vector of strings"""
    return [
        byte_vector[str_size * r : str_size * (r + 1)].decode("ASCII").strip()
        for r in range(nrows)
    ]


def pad_and_join_list_of_strings(string_list: list, str_size: int):
    """Pad each entry  to a defined size and join the padded entries into one string"""
    for i in range(len(string_list)):
        string_list[i] = string_list[i].ljust(str_size)
    result = "".join(string_list)
    return result


def numpy_array_to_ctypes(arr):
    """Cast an array to ctypes only if its len is not 0"""
    if len(arr) == 0:
        return None
    return as_ctypes(arr)


class CUGridNetwork1D(Structure):
    """C-structure intended for internal use only.
    It represents a Network1d struct as described by the UGrid API.

    Used for communicating with the UGrid dll.

    Attributes:
        name (c_char_p): The network name.
        node_x (POINTER(c_double)): The x-coordinates of the network node.
        node_y (POINTER(c_double)): The y-coordinates of the network node.
        node_id (c_char_p): The node names ids.
        node_long_name (c_char_p): The node long names.
        edge_node (POINTER(c_int)): The nodes defining each branch.
        edge_length (POINTER(c_double)): The edge lengths.
        edge_order (POINTER(c_int)): The order of the branches.
        edge_id (c_char_p): The name of the branches.
        edge_long_name (c_char_p): The long name of the branches.
        geometry_nodes_x (POINTER(c_double)): The geometry nodes x coordinates.
        geometry_nodes_y (POINTER(c_double)): The geometry nodes y coordinates.
        num_edges_geometry_nodes (POINTER(c_int)): The number of geometry nodes for each branch.
        num_geometry_nodes (c_int): The number of geometry nodes.
        num_nodes (c_int): The number of network1d nodes.
        num_edges (c_int): The number of network1d branches.
        is_spherical (c_int): 1 if the coordinates are in a spherical system, 0 otherwise .
        start_index (c_int): The start index used in arrays using indices, such as in the branch_node array.
    """

    _fields_ = [
        ("name", c_char_p),
        ("node_x", POINTER(c_double)),
        ("node_y", POINTER(c_double)),
        ("node_id", c_char_p),
        ("node_long_name", c_char_p),
        ("edge_node", POINTER(c_int)),
        ("edge_length", POINTER(c_double)),
        ("edge_order", POINTER(c_int)),
        ("edge_id", c_char_p),
        ("edge_long_name", c_char_p),
        ("geometry_nodes_x", POINTER(c_double)),
        ("geometry_nodes_y", POINTER(c_double)),
        ("num_edges_geometry_nodes", POINTER(c_int)),
        ("num_geometry_nodes", c_int),
        ("num_nodes", c_int),
        ("num_edges", c_int),
        ("is_spherical", c_int),
        ("start_index", c_int),
    ]

    @staticmethod
    def from_py_structure(
        ugrid_network1D: UGridNetwork1D, name_size: int, name_long_size: int
    ) -> CUGridNetwork1D:
        """Creates a new CMesh instance from a given Mesh2d instance.

        Args:
            ugrid_network1D (UGridNetwork1D): Class of numpy instances owning the state.

        Returns:
            CMesh2d: The created CMesh2d instance.
        """

        name_padded = ugrid_network1D.name.ljust(name_long_size)

        node_id = pad_and_join_list_of_strings(ugrid_network1D.node_id, name_size)
        node_long_name = pad_and_join_list_of_strings(
            ugrid_network1D.node_long_name, name_long_size
        )

        edge_id = pad_and_join_list_of_strings(ugrid_network1D.edge_id, name_size)
        edge_long_name = pad_and_join_list_of_strings(
            ugrid_network1D.edge_long_name, name_long_size
        )

        c_ugrid_network = CUGridNetwork1D()

        # Set the pointers
        c_ugrid_network.name = c_char_p(name_padded.encode("ASCII"))
        c_ugrid_network.node_x = numpy_array_to_ctypes(ugrid_network1D.node_x)
        c_ugrid_network.node_y = numpy_array_to_ctypes(ugrid_network1D.node_y)
        c_ugrid_network.node_id = c_char_p(node_id.encode("ASCII"))
        c_ugrid_network.node_long_name = c_char_p(node_long_name.encode("ASCII"))
        c_ugrid_network.edge_node = numpy_array_to_ctypes(ugrid_network1D.edge_node)
        c_ugrid_network.edge_length = numpy_array_to_ctypes(ugrid_network1D.edge_length)
        c_ugrid_network.edge_order = numpy_array_to_ctypes(ugrid_network1D.edge_order)
        c_ugrid_network.edge_id = c_char_p(edge_id.encode("ASCII"))
        c_ugrid_network.edge_long_name = c_char_p(edge_long_name.encode("ASCII"))
        c_ugrid_network.geometry_nodes_x = numpy_array_to_ctypes(
            ugrid_network1D.geometry_nodes_x
        )
        c_ugrid_network.geometry_nodes_y = numpy_array_to_ctypes(
            ugrid_network1D.geometry_nodes_y
        )

        # Set the sizes
        c_ugrid_network.num_geometry_nodes = ugrid_network1D.geometry_nodes_x.size
        c_ugrid_network.num_nodes = ugrid_network1D.node_x.size
        c_ugrid_network.num_edges = ugrid_network1D.edge_node.size // 2
        c_ugrid_network.is_spherical = ugrid_network1D.is_spherical
        c_ugrid_network.start_index = ugrid_network1D.start_index

        return c_ugrid_network

    def allocate_memory(self, name_size: int, name_long_size: int) -> UGridNetwork1D:
        """Allocate data according to the parameters with the "num_" prefix.
        The pointers are then set to the freshly allocated memory.
        The memory is owned by the Network1D instance which is returned by this method.

        Args:
            name_size (int): The size of the names.
            name_long_size (int): The size of the long names.

        Returns:
            UGridNetwork1D: The object owning the allocated memory.
        """

        name = " " * name_long_size
        node_x = np.empty(self.num_nodes, dtype=np.double)
        node_y = np.empty(self.num_nodes, dtype=np.double)
        node_id = " " * self.num_nodes * name_size
        node_long_name = " " * self.num_nodes * name_long_size
        edge_node = np.empty(self.num_nodes, dtype=np.int)
        edge_length = np.empty(self.num_edges, dtype=np.double)
        edge_order = np.empty(self.num_edges, dtype=np.int)
        edge_id = " " * self.num_edges * name_size
        edge_long_name = " " * self.num_edges * name_long_size
        geometry_nodes_x = np.empty(self.num_geometry_nodes, dtype=np.double)
        geometry_nodes_y = np.empty(self.num_geometry_nodes, dtype=np.double)
        num_edge_geometry_nodes = np.empty(self.num_edges, dtype=np.int)

        self.name = c_char_p(name.encode("ASCII"))
        self.node_x = numpy_array_to_ctypes(node_x)
        self.node_y = numpy_array_to_ctypes(node_y)
        self.node_id = c_char_p(node_id.encode("ASCII"))
        self.node_long_name = c_char_p(node_long_name.encode("ASCII"))
        self.edge_node = numpy_array_to_ctypes(edge_node)
        self.edge_length = numpy_array_to_ctypes(edge_length)
        self.edge_order = numpy_array_to_ctypes(edge_order)
        self.edge_id = c_char_p(edge_id.encode("ASCII"))
        self.edge_long_name = c_char_p(edge_long_name.encode("ASCII"))
        self.geometry_nodes_x = numpy_array_to_ctypes(geometry_nodes_x)
        self.geometry_nodes_y = numpy_array_to_ctypes(geometry_nodes_y)
        self.num_edge_geometry_nodes = numpy_array_to_ctypes(num_edge_geometry_nodes)

        return UGridNetwork1D(
            name=name,
            node_x=node_x,
            node_y=node_y,
            edge_node=edge_node,
            edge_length=edge_length,
            edge_order=edge_order,
            geometry_nodes_x=geometry_nodes_x,
            geometry_nodes_y=geometry_nodes_y,
            num_edge_geometry_nodes=num_edge_geometry_nodes,
        )


class CUGridMesh1D(Structure):
    """C-structure intended for internal use only.
    It represents a Mesh1d struct as described by the UGrid API.

    Used for communicating with the UGrid dll.

    Attributes:
        name (c_char_p): The mesh name.
        network_name (c_char_p): The x-coordinates of the network node.
        node_x (POINTER(c_double)):  The node x coordinate.
        node_y (POINTER(c_double)):  The node y coordinate.
        edge_node (POINTER(c_int)): The edge node connectivity.
        node_edge_id (POINTER(c_int)):  The network edge id where every node lies.
        branch_offset (POINTER(c_double)): The offset of each node on the network edge
        node_name_id (POINTER(c_int)): The node name.
        node_name_long (c_char_p): The node long name.
        edge_edge_id (c_char_p): The network edge id where every edge lies.
        edge_edge_offset (POINTER(c_double)): The offset of each edge on the network edge.
        edge_x (POINTER(c_double)): The edge x coordinate.
        edge_y (c_int): The edge y coordinate.
        num_nodes (c_int): The number of nodes.
        num_edges (c_int): The number of edges.
        is_spherical (c_int): 1 if coordinates are in a spherical system, 0 otherwise.
        start_index (c_int): The start index used in arrays using indices, such as in the branch_node array.
        double_fill_value (c_double): The fill value for array of doubles.
        int_fill_value (c_int): The fill value for array of integers.
    """

    _fields_ = [
        ("name", c_char_p),
        ("network_name", c_char_p),
        ("node_x", POINTER(c_double)),
        ("node_y", POINTER(c_double)),
        ("edge_node", POINTER(c_int)),
        ("node_edge_id", POINTER(c_int)),
        ("branch_offset", POINTER(c_double)),
        ("node_name_id", c_char_p),
        ("node_name_long", c_char_p),
        ("edge_edge_id", POINTER(c_int)),
        ("edge_edge_offset", POINTER(c_double)),
        ("edge_x", POINTER(c_double)),
        ("edge_y", POINTER(c_double)),
        ("num_nodes", c_int),
        ("num_edges", c_int),
        ("is_spherical", c_int),
        ("start_index", c_int),
        ("double_fill_value", c_double),
        ("int_fill_value", c_int),
    ]

    @staticmethod
    def from_py_structure(
        mesh1d: UGridMesh1D, name_size: int, name_long_size: int
    ) -> CUGridMesh1D:
        """Creates a new CMesh instance from a given Mesh2d instance.

        Args:
            name_size (int): The size of the names.
            name_long_size (int): The size of the long names.
            mesh1d (UGridMesh1D): Class of numpy instances owning the state.

        Returns:
            CMesh2d: The created CMesh2d instance.
        """

        c_mesh1d = CUGridMesh1D()

        # Set the pointers
        mesh1d_name_padded = mesh1d.name.ljust(name_long_size)
        network1d_name_padded = mesh1d.network_name.ljust(name_long_size)
        node_name_id = pad_and_join_list_of_strings(mesh1d.node_name_id, name_size)
        node_name_long = pad_and_join_list_of_strings(
            mesh1d.node_name_long, name_long_size
        )

        c_mesh1d.name = c_char_p(mesh1d_name_padded.encode("ASCII"))
        c_mesh1d.network_name = c_char_p(network1d_name_padded.encode("ASCII"))
        c_mesh1d.node_x = numpy_array_to_ctypes(mesh1d.node_x)
        c_mesh1d.node_y = numpy_array_to_ctypes(mesh1d.node_y)
        c_mesh1d.edge_node = numpy_array_to_ctypes(mesh1d.edge_node)
        c_mesh1d.node_edge_id = numpy_array_to_ctypes(mesh1d.node_edge_id)
        c_mesh1d.branch_offset = numpy_array_to_ctypes(mesh1d.node_edge_offset)
        c_mesh1d.node_name_id = c_char_p(node_name_id.encode("ASCII"))
        c_mesh1d.node_name_long = c_char_p(node_name_long.encode("ASCII"))
        c_mesh1d.edge_edge_id = numpy_array_to_ctypes(mesh1d.edge_edge_id)
        c_mesh1d.edge_edge_offset = numpy_array_to_ctypes(mesh1d.edge_edge_offset)
        c_mesh1d.edge_x = numpy_array_to_ctypes(mesh1d.edge_x)
        c_mesh1d.edge_y = numpy_array_to_ctypes(mesh1d.edge_y)

        # Set the sizes
        c_mesh1d.num_nodes = mesh1d.node_x.size
        c_mesh1d.num_edges = mesh1d.edge_node.size // 2

        # Set other properties
        c_mesh1d.is_spherical = mesh1d.is_spherical
        c_mesh1d.start_index = mesh1d.start_index

        return c_mesh1d

    def allocate_memory(self, name_size: int, name_long_size: int) -> UGridMesh1D:
        """Allocate data according to the parameters with the "num_" prefix.
        The pointers are then set to the freshly allocated memory.
        The memory is owned by the Mesh1d instance which is returned by this method.

        Args:
            name_size (int): The size of the names.
            name_long_size (int): The size of the long names.

        Returns:
            UGridMesh1D: The object owning the allocated memory.
        """

        name = " " * name_long_size
        network_name = " " * name_long_size
        node_x = np.empty(self.num_nodes, dtype=np.double)
        node_y = np.empty(self.num_nodes, dtype=np.double)
        edge_node = np.empty(self.num_edges * 2, dtype=np.int)
        branch_id = np.empty(self.num_nodes, dtype=np.int)
        branch_offset = np.empty(self.num_nodes, dtype=np.double)
        node_name_id = " " * self.num_nodes * name_size
        node_name_long = " " * self.num_nodes * name_long_size
        edge_edge_id = np.empty(self.num_edges, dtype=np.int)
        edge_edge_offset = np.empty(self.num_edges, dtype=np.double)
        edge_x = np.empty(self.num_edges, dtype=np.double)
        edge_y = np.empty(self.num_edges, dtype=np.double)

        self.name = c_char_p(name.encode("ASCII"))
        self.network_name = c_char_p(network_name.encode("ASCII"))
        self.node_x = numpy_array_to_ctypes(node_x)
        self.node_y = numpy_array_to_ctypes(node_y)
        self.edge_node = numpy_array_to_ctypes(edge_node)
        self.node_edge_id = numpy_array_to_ctypes(branch_id)
        self.branch_offset = numpy_array_to_ctypes(branch_offset)
        self.node_name_id = c_char_p(node_name_id.encode("ASCII"))
        self.node_name_long = c_char_p(node_name_long.encode("ASCII"))
        self.edge_edge_id = numpy_array_to_ctypes(edge_edge_id)
        self.edge_edge_offset = numpy_array_to_ctypes(edge_edge_offset)
        self.edge_x = numpy_array_to_ctypes(edge_x)
        self.edge_y = numpy_array_to_ctypes(edge_y)

        return UGridMesh1D(
            name=name,
            network_name=network_name,
            node_edge_id=branch_id,
            node_edge_offset=branch_offset,
            node_x=node_x,
            node_y=node_y,
            edge_node=edge_node,
            edge_edge_id=edge_edge_id,
            edge_edge_offset=edge_edge_offset,
            edge_x=edge_x,
            edge_y=edge_y,
            node_name_id=node_name_id,
            node_name_long=node_name_long,
        )


class CUGridMesh2D(Structure):
    """C-structure intended for internal use only.
    It represents a Mesh2D struct as described by the UGrid API.

    Used for communicating with the UGrid dll.

    Attributes:
        name (c_char_p): The mesh name.
        edge_nodes (POINTER(c_int)): The nodes composing each mesh 2d edge.
        face_nodes (POINTER(c_int)): The nodes composing each mesh 2d face.
        nodes_per_face (POINTER(c_int)): The nodes composing each mesh 2d face.
        node_x (POINTER(c_double)): The x-coordinates of the nodes.
        node_y (POINTER(c_double)): The y-coordinates of the nodes.
        edge_x (POINTER(c_double)): The x-coordinates of the mesh edges' middle points.
        edge_y (POINTER(c_double)): The x-coordinates of the mesh edges' middle points.
        face_x (POINTER(c_double)): The x-coordinates of the mesh faces' mass centers.
        face_y (POINTER(c_double)): The y-coordinates of the mesh faces' mass centers.
        edge_face (POINTER(c_int)): The edges composing each face.
        face_edge (POINTER(c_int)): For each face, the edges composing it.
        face_face (POINTER(c_int)): For each face, the neighboring faces.
        node_z (POINTER(c_double)): The node z coordinates.
        edge_z (POINTER(c_double)): The edge z coordinates.
        face_z (POINTER(c_double)): The face z coordinates.
        layer_zs (POINTER(c_double)): The z coordinates of a layer.
        interface_zs (POINTER(c_double)): The z coordinates of a layer interface.
        boundary_node_connectivity (POINTER(c_double)): To be detailed.
        volume_coordinates (POINTER(c_double)): To be detailed.
        num_nodes (c_int): The number of mesh nodes.
        num_edges (c_int): The number of edges.
        num_faces (c_int): The number of faces.
        num_layers (c_int): The number of layers.
        start_index (c_int): The start index used in arrays using indices, such as in the branch_node array.
        num_face_nodes_max (c_int): 1 if coordinates are in a spherical system, 0 otherwise.
        is_spherical (c_int): 1 if coordinates are in a spherical system, 0 otherwise.
        double_fill_value (c_double): The fill value for array of doubles.
        int_fill_value (c_int): The fill value for array of integers.
    """

    _fields_ = [
        ("name", c_char_p),
        ("edge_node", POINTER(c_int)),
        ("face_node", POINTER(c_int)),
        ("node_x", POINTER(c_double)),
        ("node_y", POINTER(c_double)),
        ("edge_x", POINTER(c_double)),
        ("edge_y", POINTER(c_double)),
        ("face_x", POINTER(c_double)),
        ("face_y", POINTER(c_double)),
        ("edge_face", POINTER(c_int)),
        ("face_edge", POINTER(c_int)),
        ("face_face", POINTER(c_int)),
        ("node_z", POINTER(c_double)),
        ("edge_z", POINTER(c_double)),
        ("face_z", POINTER(c_double)),
        ("layer_zs", POINTER(c_double)),
        ("interface_zs", POINTER(c_double)),
        ("boundary_node_connectivity", POINTER(c_double)),
        ("volume_coordinates", POINTER(c_int)),
        ("num_nodes", c_int),
        ("num_edges", c_int),
        ("num_faces", c_int),
        ("num_layers", c_int),
        ("start_index", c_int),
        ("num_face_nodes_max", c_int),
        ("is_spherical", c_int),
        ("double_fill_value", c_double),
        ("int_fill_value", c_int),
    ]

    @staticmethod
    def from_py_structure(mesh2d: UGridMesh2D, name_long_size: int) -> CUGridMesh2D:
        """Creates a new CMesh instance from a given Mesh2d instance.

        Args:
            mesh2d (UGridMesh2D): Class of numpy instances owning the state.
            name_long_size (int): The size of the long names.

        Returns:
            CMesh2d: The created CMesh2d instance.
        """

        c_mesh2d = CUGridMesh2D()

        # Required arrays
        mesh2d_name_padded = mesh2d.name.ljust(name_long_size)
        c_mesh2d.name = c_char_p(mesh2d_name_padded.encode("ASCII"))
        c_mesh2d.edge_node = numpy_array_to_ctypes(mesh2d.edge_node)
        c_mesh2d.node_x = numpy_array_to_ctypes(mesh2d.node_x)
        c_mesh2d.node_y = numpy_array_to_ctypes(mesh2d.node_y)

        # Optional arrays
        c_mesh2d.edge_x = numpy_array_to_ctypes(mesh2d.edge_x)
        c_mesh2d.edge_y = numpy_array_to_ctypes(mesh2d.edge_y)
        c_mesh2d.face_x = numpy_array_to_ctypes(mesh2d.face_x)
        c_mesh2d.face_y = numpy_array_to_ctypes(mesh2d.face_y)
        c_mesh2d.edge_face = numpy_array_to_ctypes(mesh2d.edge_face)

        c_mesh2d.face_node = numpy_array_to_ctypes(mesh2d.face_node)
        c_mesh2d.face_edge = numpy_array_to_ctypes(mesh2d.face_edge)
        c_mesh2d.face_face = numpy_array_to_ctypes(mesh2d.face_face)

        c_mesh2d.node_z = numpy_array_to_ctypes(mesh2d.node_z)
        c_mesh2d.edge_z = numpy_array_to_ctypes(mesh2d.edge_z)
        c_mesh2d.face_z = numpy_array_to_ctypes(mesh2d.face_z)
        c_mesh2d.layer_zs = numpy_array_to_ctypes(mesh2d.layer_zs)
        c_mesh2d.interface_zs = numpy_array_to_ctypes(mesh2d.interface_zs)
        c_mesh2d.boundary_node_connectivity = numpy_array_to_ctypes(
            mesh2d.boundary_node_connectivity
        )
        c_mesh2d.volume_coordinates = numpy_array_to_ctypes(mesh2d.volume_coordinates)

        # Set the sizes
        c_mesh2d.num_nodes = mesh2d.node_x.size
        c_mesh2d.num_edges = mesh2d.edge_node.size // 2
        c_mesh2d.num_faces = mesh2d.face_x.size
        c_mesh2d.num_layers = mesh2d.layer_zs.size

        # Set other properties
        c_mesh2d.start_index = mesh2d.start_index
        c_mesh2d.num_face_nodes_max = mesh2d.num_face_nodes_max
        c_mesh2d.is_spherical = mesh2d.is_spherical
        c_mesh2d.double_fill_value = mesh2d.double_fill_value
        c_mesh2d.int_fill_value = mesh2d.int_fill_value

        return c_mesh2d

    def allocate_memory(self, name_long_size: int) -> UGridMesh2D:
        """Allocate data according to the parameters with the "num_" prefix.
        The pointers are then set to the freshly allocated memory.
        The memory is owned by the Mesh2D instance which is returned by this method.

        Args:
            name_long_size (int): The size of the long names.

        Returns:
            UGridMesh2D: The object owning the allocated memory.
        """

        name = " " * name_long_size
        node_x = np.empty(self.num_nodes, dtype=np.double)
        node_y = np.empty(self.num_nodes, dtype=np.double)
        edge_node = np.empty(self.num_edges * 2, dtype=np.int)

        node_z = np.empty(self.num_nodes, dtype=np.double)
        edge_x = np.empty(self.num_edges, dtype=np.double)
        edge_y = np.empty(self.num_edges, dtype=np.double)
        edge_z = np.empty(self.num_edges, dtype=np.double)
        edge_face = np.empty(self.num_edges * 2, dtype=np.int)

        face_node = np.empty(self.num_faces * self.num_face_nodes_max, dtype=np.int)
        face_edge = np.empty(self.num_faces * self.num_face_nodes_max, dtype=np.int)
        face_face = np.empty(self.num_faces * self.num_face_nodes_max, dtype=np.int)
        face_x = np.empty(self.num_faces, dtype=np.double)
        face_y = np.empty(self.num_faces, dtype=np.double)
        face_z = np.empty(self.num_faces, dtype=np.double)

        layer_zs = np.empty(self.num_layers, dtype=np.double)
        boundary_node_connectivity = np.empty(self.num_nodes, dtype=np.double)
        volume_coordinates = np.empty(self.num_faces, dtype=np.int)

        self.name = c_char_p(name.encode("ASCII"))
        self.edge_node = numpy_array_to_ctypes(edge_node)
        self.face_node = numpy_array_to_ctypes(face_node)
        self.node_x = numpy_array_to_ctypes(node_x)
        self.node_y = numpy_array_to_ctypes(node_y)
        self.edge_x = numpy_array_to_ctypes(edge_x)
        self.edge_y = numpy_array_to_ctypes(edge_y)
        self.face_x = numpy_array_to_ctypes(face_x)
        self.face_y = numpy_array_to_ctypes(face_y)
        self.edge_face = numpy_array_to_ctypes(edge_face)
        self.face_edge = numpy_array_to_ctypes(face_edge)
        self.face_face = numpy_array_to_ctypes(face_face)
        self.node_z = numpy_array_to_ctypes(node_z)
        self.edge_z = numpy_array_to_ctypes(edge_z)
        self.face_z = numpy_array_to_ctypes(face_z)

        self.layer_zs = numpy_array_to_ctypes(layer_zs)
        self.boundary_node_connectivity = numpy_array_to_ctypes(
            boundary_node_connectivity
        )
        self.volume_coordinates = numpy_array_to_ctypes(volume_coordinates)

        if self.num_layers > 1:
            interface_zs = np.empty(self.num_layers - 1, dtype=np.double)
            self.interface_zs = numpy_array_to_ctypes(interface_zs)
        else:
            interface_zs = None
            self.interface_zs = None

        return UGridMesh2D(
            name=name,
            node_x=node_x,
            node_y=node_y,
            edge_node=edge_node,
            face_node=face_node,
            edge_x=edge_x,
            edge_y=edge_y,
            face_x=face_x,
            face_y=face_y,
            edge_face=edge_face,
            face_edge=face_edge,
            face_face=face_face,
            node_z=node_z,
            edge_z=edge_z,
            face_z=face_z,
            layer_zs=layer_zs,
            interface_zs=interface_zs,
            boundary_node_connectivity=boundary_node_connectivity,
            volume_coordinates=volume_coordinates,
        )


class CUGridContacts(Structure):
    """C-structure intended for internal use only.
    It represents a Contacts struct as described by the UGrid API.

    Used for communicating with the UGrid dll.

    Attributes:
        name (c_char_p): The name of the contact entity.
        edges (POINTER(c_int)): The actual contacts, expressed as pair of indices from a mesh index
        to another mesh index.
        contact_type (POINTER(c_int)): For each contact its type.
        contact_name_id (c_char_p): The name of each contact.
        contact_name_long (c_char_p): The long name of each contact.
        mesh_from_name (c_char_p): The name of the mesh where the contacts start.
        mesh_to_name (c_char_p): The name of the mesh where the contacts ends.
        mesh_from_location (c_int): The location type (node, edge or face) at the contact start.
        mesh_to_location (c_int): The location type (node, edge or face) at the contact end.
        num_contacts (c_int): The number of contacts.
    """

    _fields_ = [
        ("name", c_char_p),
        ("edges", POINTER(c_int)),
        ("contact_type", POINTER(c_int)),
        ("contact_name_id", c_char_p),
        ("contact_name_long", c_char_p),
        ("mesh_from_name", c_char_p),
        ("mesh_to_name", c_char_p),
        ("mesh_from_location", c_int),
        ("mesh_to_location", c_int),
        ("num_contacts", c_int),
    ]

    @staticmethod
    def from_py_structure(
        contacts: UGridContacts, name_size: int, name_long_size: int
    ) -> CUGridContacts:
        """Creates a new CContacts instance from a given Contacts instance.

        Args:
            contacts (UGridContacts): Class of numpy instances owning the state.

        Returns:
            CUGridContacts: The created CContacts instance.
        """

        c_contacts = CUGridContacts()

        # Set the pointers
        contacts_name_padded = contacts.name.ljust(name_long_size)
        mesh_from_name_padded = contacts.mesh_from_name.ljust(name_long_size)
        mesh_to_name_padded = contacts.mesh_to_name.ljust(name_long_size)

        contact_name_id_padded = pad_and_join_list_of_strings(
            contacts.contact_name_id, name_size
        )
        contact_name_long_padded = pad_and_join_list_of_strings(
            contacts.contact_name_long, name_long_size
        )

        c_contacts.name = c_char_p(contacts_name_padded.encode("ASCII"))
        c_contacts.edges = numpy_array_to_ctypes(contacts.edges)
        c_contacts.contact_type = numpy_array_to_ctypes(contacts.contact_type)
        c_contacts.mesh_from_name = c_char_p(mesh_from_name_padded.encode("ASCII"))
        c_contacts.mesh_to_name = c_char_p(mesh_to_name_padded.encode("ASCII"))
        c_contacts.contact_name_id = c_char_p(contact_name_id_padded.encode("ASCII"))
        c_contacts.contact_name_long = c_char_p(
            contact_name_long_padded.encode("ASCII")
        )
        c_contacts.mesh_from_location = contacts.mesh_from_location
        c_contacts.mesh_to_location = contacts.mesh_to_location

        # Set the size
        c_contacts.num_contacts = contacts.edges.size // 2

        return c_contacts

    def allocate_memory(self, name_size: int, name_long_size: int) -> UGridContacts:
        """Allocate data according to the parameters with the "num_" prefix.
        The pointers are then set to the freshly allocated memory.
        The memory is owned by the Contacts instance which is returned by this method.

        Args:
            name_long_size (int): The size of the long names.

        Returns:
            UGridContacts: The object owning the allocated memory.
        """

        name = " " * name_long_size
        mesh_from_name = " " * name_long_size
        mesh_to_name = " " * name_long_size
        edges = np.empty(self.num_contacts * 2, dtype=np.int)
        contact_type = np.empty(self.num_contacts, dtype=np.int)
        contact_name_id = " " * self.num_contacts * name_size
        contact_name_long = " " * self.num_contacts * name_long_size

        self.name = c_char_p(name.encode("ASCII"))
        self.edges = numpy_array_to_ctypes(edges)
        self.contact_type = numpy_array_to_ctypes(contact_type)
        self.contact_name_id = c_char_p(contact_name_id.encode("ASCII"))
        self.contact_name_long = c_char_p(contact_name_long.encode("ASCII"))
        self.mesh_from_name = c_char_p(mesh_from_name.encode("ASCII"))
        self.mesh_to_name = c_char_p(mesh_to_name.encode("ASCII"))

        return UGridContacts(
            name=name,
            edges=edges,
            mesh_from_name=mesh_from_name,
            mesh_to_name=mesh_to_name,
            contact_type=contact_type,
            contact_name_id=contact_name_id,
            contact_name_long=contact_name_long,
        )
